from django.contrib import admin
from django.template.loader import render_to_string
from django_audit_fields.admin import audit_fieldset_tuple
from edc_consent import site_consents
from edc_constants.constants import NO, YES
from edc_model_admin.mixins import ModelAdminProtectPiiMixin
from edc_sites import site_sites
from edc_utils import age, get_utcnow

from .fieldsets import (
    indirect_contacts_fieldset,
    subject_contacts_fieldset,
    work_contacts_fieldset,
)
from .forms import SubjectLocatorForm


class SubjectLocatorModelAdminMixin(ModelAdminProtectPiiMixin):
    form = SubjectLocatorForm

    extra_pii_attrs: list[str] = ["contacts"]

    fieldsets = (
        (None, {"fields": ("subject_identifier",)}),
        subject_contacts_fieldset,
        work_contacts_fieldset,
        indirect_contacts_fieldset,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "may_visit_home": admin.VERTICAL,
        "may_call": admin.VERTICAL,
        "may_sms": admin.VERTICAL,
        "may_call_work": admin.VERTICAL,
        "may_contact_indirectly": admin.VERTICAL,
    }

    list_filter = (
        "may_visit_home",
        "may_call",
        "may_sms",
        "may_call_work",
        "may_contact_indirectly",
    )

    list_display = (
        "subject",
        "dashboard",
        "contacts",
        "contact_rules",
    )

    search_fields = (
        "subject_identifier",
        "subject_cell__exact",
        "subject_cell_alt__exact",
        "subject_phone__exact",
        "subject_phone_alt__exact",
        "subject_work_phone__exact",
        "subject_work_cell__exact",
        "indirect_contact_cell__exact",
        "indirect_contact_cell_alt__exact",
        "indirect_contact_phone__exact",
    )

    @admin.display(description="Subject", ordering="subject_identifier")
    def subject(self, obj):
        cdef = site_consents.get_consent_definition(
            report_datetime=obj.report_datetime, site=site_sites.get(obj.site.id)
        )
        consent = cdef.model_cls.objects.get(subject_identifier=obj.subject_identifier)
        context = dict(
            subject_identifier=obj.subject_identifier,
            gender=consent.gender.upper(),
            age_in_years=age(born=consent.dob, reference_dt=get_utcnow()).years,
            initials=consent.initials,
        )
        return render_to_string("changelist_locator_subject.html", context=context)

    @admin.display(description="Contact Rules", ordering="may_call")
    def contact_rules(self, obj):
        context = dict(
            may_visit_home=obj.may_visit_home,
            may_call=obj.may_call,
            sms=obj.sms,
            call_work=obj.call_work,
            contact_indirectly=obj.contact_indirectly,
            YES=YES,
            NO=NO,
        )
        return render_to_string("changelist_locator_contact_rules.html", context=context)

    @admin.display(description="Contacts")
    def contacts(self, obj):
        context = dict(
            subject_cell=obj.subject_cell,
            subject_cell_alt=obj.subject_cell_alt,
            subject_phone=obj.subject_phone,
            subject_phone_alt=obj.subject_phone_alt,
        )
        return render_to_string("changelist_locator_contacts.html", context=context)



from django.db import models
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_utils import get_utcnow

from .locator_methods_model_mixin import LocatorMethodsModelMixin
from .subject_contact_fields_mixin import SubjectContactFieldsMixin
from .subject_indirect_contact_fields_mixin import SubjectIndirectContactFieldsMixin
from .subject_work_fields_mixin import SubjectWorkFieldsMixin


class LocatorManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, subject_identifier):
        return self.get(subject_identifier=subject_identifier)


class LocatorModelMixin(
    UniqueSubjectIdentifierFieldMixin,
    SubjectContactFieldsMixin,
    SubjectIndirectContactFieldsMixin,
    SubjectWorkFieldsMixin,
    LocatorMethodsModelMixin,
    models.Model,
):
    """A model completed by the user to that captures participant
    locator information and permission to contact.

    This model is used in ListboardViews.
    """

    report_datetime = models.DateTimeField(default=get_utcnow)

    objects = LocatorManager()

    def __str__(self):
        return self.subject_identifier

    def natural_key(self):
        return (self.subject_identifier,)

    class Meta:
        abstract = True



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings

if TYPE_CHECKING:
    from .model_mixins import LocatorModelMixin


class LocatorModelError(Exception):
    pass


def get_locator_model() -> str:
    return getattr(settings, "SUBJECT_LOCATOR_MODEL", "edc_locator.subjectlocator")


def get_locator_model_cls() -> LocatorModelMixin:
    return django_apps.get_model(get_locator_model())



# Generated by Django 4.2.1 on 2023-07-07 19:32

import edc_sites.models
from django.db import migrations

import edc_locator.model_mixins.locator_model_mixin


class Migration(migrations.Migration):
    dependencies = [
        ("edc_locator", "0031_alter_subjectlocator_options"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="subjectlocator",
            managers=[
                (
                    "objects",
                    edc_locator.model_mixins.locator_model_mixin.LocatorManager(),
                ),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



# Generated by Django 2.1 on 2018-10-06 22:53

import django.db.models.deletion
import edc_sites.models
from django.db import migrations, models

import edc_locator.model_mixins.locator_model_mixin


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0011_auto_20181009_2236"),
        ("edc_locator", "0010_auto_20180809_0301"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="subjectlocator",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                (
                    "objects",
                    edc_locator.model_mixins.locator_model_mixin.LocatorManager(),
                ),
            ],
        ),
        migrations.RenameField(
            model_name="historicalsubjectlocator",
            old_name="parent_reference_identifier",
            new_name="parent_action_identifier",
        ),
        migrations.RenameField(
            model_name="historicalsubjectlocator",
            old_name="related_reference_identifier",
            new_name="related_action_identifier",
        ),
        migrations.RenameField(
            model_name="subjectlocator",
            old_name="parent_reference_identifier",
            new_name="parent_action_identifier",
        ),
        migrations.RenameField(
            model_name="subjectlocator",
            old_name="related_reference_identifier",
            new_name="related_action_identifier",
        ),
        migrations.AddField(
            model_name="historicalsubjectlocator",
            name="action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="subjectlocator",
            name="action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="action_identifier",
            field=models.CharField(null=True, max_length=50),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="action_identifier",
            field=models.CharField(max_length=50, null=True),
        ),
    ]



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0010_auto_20181009_0445"),
        ("edc_locator", "0011_auto_20181007_0053"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalsubjectlocator",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectlocator",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="subjectlocator",
            name="parent_action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="subjectlocator",
            name="related_action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
    ]



from django import forms
from edc_action_item.forms import ActionItemFormMixin
from edc_form_validators import FormValidatorMixin

from ..models import SubjectLocator
from .subject_locator_form_validator import SubjectLocatorFormValidator


class SubjectLocatorForm(FormValidatorMixin, ActionItemFormMixin, forms.ModelForm):
    form_validator_cls = SubjectLocatorFormValidator

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = SubjectLocator
        fields = "__all__"



from edc_action_item import Action, site_action_items
from edc_constants.constants import HIGH_PRIORITY

from edc_locator.utils import get_locator_model

SUBJECT_LOCATOR_ACTION = "submit-subject-locator"


# TODO: reference model name may not match that specified in visit schedule??
class SubjectLocatorAction(Action):
    name = SUBJECT_LOCATOR_ACTION
    display_name = "Submit Subject Locator"
    reference_model = get_locator_model()
    show_link_to_changelist = True
    admin_site_name = "edc_locator_admin"
    priority = HIGH_PRIORITY
    singleton = True


site_action_items.register(SubjectLocatorAction)



# Generated by Django 2.0.1 on 2018-01-17 16:19

import edc_sites.models
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0006_auto_20180116_1528")]

    operations = [
        migrations.AlterModelManagers(
            name="subjectlocator",
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        )
    ]



from unittest.case import skip

from django.test import TestCase
from django.views.generic.base import ContextMixin
from edc_action_item import site_action_items
from edc_action_item.models import ActionItem
from edc_registration.models import RegisteredSubject
from edc_sites.view_mixins import SiteViewMixin
from edc_subject_dashboard.view_mixins import RegisteredSubjectViewMixin
from edc_test_utils.get_httprequest_for_tests import get_request_object_for_tests
from edc_test_utils.get_user_for_tests import get_user_for_tests

from edc_locator.action_items import SUBJECT_LOCATOR_ACTION
from edc_locator.exceptions import SubjectLocatorViewMixinError
from edc_locator.view_mixins import SubjectLocatorViewMixin


class TestViewMixins(TestCase):
    def setUp(self):
        self.user = get_user_for_tests()
        self.subject_identifier = "12345"
        RegisteredSubject.objects.create(subject_identifier=self.subject_identifier)

    def test_subject_locator_raises_on_bad_model(self):
        class MySubjectLocatorViewMixin(
            SiteViewMixin, SubjectLocatorViewMixin, RegisteredSubjectViewMixin, ContextMixin
        ):
            subject_locator_model = "blah.blahblah"

        mixin = MySubjectLocatorViewMixin()
        mixin.kwargs = {"subject_identifier": self.subject_identifier}
        mixin.request = get_request_object_for_tests(self.user)
        self.assertRaises(LookupError, mixin.get_context_data)

    @skip("problems emulating message framework")
    def test_mixin_messages(self):
        class MySubjectLocatorViewMixin(
            SiteViewMixin, SubjectLocatorViewMixin, RegisteredSubjectViewMixin, ContextMixin
        ):
            subject_locator_model = "edc_locator.subjectlocator"

        mixin = MySubjectLocatorViewMixin()
        mixin.kwargs = {"subject_identifier": self.subject_identifier}
        mixin.request = get_request_object_for_tests(self.user)
        self.assertGreater(len(mixin.request._messages._queued_messages), 0)

    def test_subject_locator_view_ok(self):
        class MySubjectLocatorViewMixin(
            SiteViewMixin, SubjectLocatorViewMixin, RegisteredSubjectViewMixin, ContextMixin
        ):
            subject_locator_model = "edc_locator.subjectlocator"

        mixin = MySubjectLocatorViewMixin()
        mixin.request = get_request_object_for_tests(self.user)
        mixin.kwargs = {"subject_identifier": self.subject_identifier}
        try:
            mixin.get_context_data()
        except SubjectLocatorViewMixinError as e:
            self.fail(e)

    def test_subject_locator_self_corrects_if_multiple_actionitems(self):
        class MySubjectLocatorViewMixin(
            SiteViewMixin, SubjectLocatorViewMixin, RegisteredSubjectViewMixin, ContextMixin
        ):
            subject_locator_model = "edc_locator.subjectlocator"

        mixin = MySubjectLocatorViewMixin()
        mixin.request = get_request_object_for_tests(self.user)
        mixin.kwargs = {"subject_identifier": self.subject_identifier}
        try:
            mixin.get_context_data()
        except SubjectLocatorViewMixinError as e:
            self.fail(e)
        action_cls = site_action_items.get(SUBJECT_LOCATOR_ACTION)
        action_item_model_cls = action_cls.action_item_model_cls()
        action_cls(subject_identifier=self.subject_identifier)
        action_item = ActionItem.objects.get(subject_identifier=self.subject_identifier)
        self.assertEqual(action_item_model_cls.objects.all().count(), 1)
        action_item.subject_identifier = f"{self.subject_identifier}-bad"
        action_item.save()
        self.assertEqual(action_item_model_cls.objects.all().count(), 1)
        action_cls = site_action_items.get(SUBJECT_LOCATOR_ACTION)
        action_cls(subject_identifier=self.subject_identifier)
        action_item.subject_identifier = self.subject_identifier
        action_item.save()
        self.assertEqual(action_item_model_cls.objects.all().count(), 2)
        try:
            mixin.get_context_data()
        except SubjectLocatorViewMixinError as e:
            self.fail(e)
        self.assertEqual(action_item_model_cls.objects.all().count(), 1)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

from django.apps import apps as django_apps
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from edc_action_item.site_action_items import site_action_items

from ..action_items import SUBJECT_LOCATOR_ACTION
from ..exceptions import SubjectLocatorViewMixinError
from ..utils import get_locator_model

if TYPE_CHECKING:
    from ..models import SubjectLocator


class SubjectLocatorViewMixin:
    """Adds subject locator to the context.

    Declare with RegisteredSubjectViewMixin.
    """

    subject_locator_model: int = get_locator_model()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.subject_locator_model:
            raise SubjectLocatorViewMixinError(
                "subject_locator_model must be a model (label_lower). Got None"
            )

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        subject_locator = self.subject_locator
        if not subject_locator:
            subject_locator = self.subject_locator_model_cls(
                subject_identifier=self.subject_identifier
            )
            self.create_subject_locator_action()
        kwargs.update(subject_locator=subject_locator)
        return super().get_context_data(**kwargs)

    def create_subject_locator_action(self) -> None:
        """Create a subject locator action"""

        # kwargs `subject_identifier` updated by RegisteredSubject
        # view mixin get()
        subject_identifier = self.kwargs.get("subject_identifier")
        action_cls = site_action_items.get(SUBJECT_LOCATOR_ACTION)
        action_item_model_cls = action_cls.action_item_model_cls()
        try:
            action_item_model_cls.objects.get(
                subject_identifier=subject_identifier,
                action_type__name=SUBJECT_LOCATOR_ACTION,
            )
        except ObjectDoesNotExist:
            # only create missing action item if user has change perms
            app_label, model_name = self.subject_locator_model_cls._meta.label_lower.split(".")
            if self.request.user.has_perm(f"{app_label}.change_{model_name}"):
                action_cls(subject_identifier=subject_identifier)
        except MultipleObjectsReturned:
            # if condition exists, correct here
            action_item_model_cls.objects.filter(
                subject_identifier=subject_identifier,
                action_type__name=SUBJECT_LOCATOR_ACTION,
            ).delete()
            action_cls(subject_identifier=subject_identifier)

    @property
    def subject_locator_model_cls(self) -> Type[SubjectLocator]:
        return django_apps.get_model(self.subject_locator_model)

    @property
    def subject_locator(self) -> SubjectLocator | None:
        """Returns a model instance or None"""
        try:
            obj = self.subject_locator_model_cls.objects.get(
                subject_identifier=self.subject_identifier
            )
        except ObjectDoesNotExist:
            obj = None
        return obj
